import { forArea } from 'complog';
import { __extends } from 'tslib';
import * as tslib_1 from 'tslib';

var appLogger = forArea("t-rex");

var ROOT = "/";
var $$MODIFIED = "$$t-rex:modified";
var $$VERSION = "$$t-rex:version";
var logger$2 = appLogger.create("TransactionalObject");
var TransactionalObject = (function () {
    function TransactionalObject(initialState) {
        this.base = this.current = initialState;
        this.modified = [];
    }
    TransactionalObject.prototype.commit = function () {
        logger$2("commit").log();
        for (var _i = 0, _a = this.modified; _i < _a.length; _i++) {
            var obj = _a[_i];
            delete obj[$$MODIFIED];
        }
        this.modified = [];
        this.base = this.current;
    };
    TransactionalObject.prototype.rebase = function (newBase) {
        logger$2("rebase").log();
        this.current = this.internalRebase(this.current, this.base, newBase, "/");
        this.base = newBase;
    };
    TransactionalObject.prototype.appendPath = function (path, field) {
        if (path == "/") {
            return "/" + field;
        }
        return path + "/" + field;
    };
    TransactionalObject.prototype.internalRebase = function (current, base, newBase, path) {
        if (newBase == base) {
            return current;
        }
        if (newBase[$$VERSION] < base[$$VERSION]) {
            return current;
        }
        var newCurrent = current;
        for (var field in newBase) {
            if (field.startsWith("$$t-rex")) {
                continue;
            }
            if (base[field] === current[field]) {
                //
                //  We hold no change for this field
                //  Any new change is welcomed
                //
                newCurrent = this.setField(current, field, newBase[field]);
                continue;
            }
            if (typeof base[field] !== "object") {
                //
                //  There is a local change inside a value type
                //  If newBase has different value than this is a conflict
                //
                if (base[field] != newBase[field]) {
                    var fieldPath = this.appendPath(path, field);
                    logger$2("Concurrency error at path \"" + fieldPath + "\".", "base is", base[field], "while latest is", newBase[field]).error();
                    throw new Error("Concurrency error at " + fieldPath);
                }
            }
            else {
                //
                //  There is a local change inside a reference type
                //  If newBase is older than our data than we are OK
                //
                if (base[field] === newBase[field] || base[field][$$VERSION] > newBase[field][$$VERSION]) {
                    continue;
                }
                //
                //  newBase is same version or bigger
                //  This implies a chance that newBase holds new values that might be conflicting
                //
                var newVal = this.internalRebase(current[field], base[field], newBase[field], (path == "/" ? path + field : path + "/" + field));
                newCurrent = this.setField(current, field, newVal);
            }
        }
        return newCurrent;
    };
    TransactionalObject.prototype.getBase = function () {
        return this.base;
    };
    TransactionalObject.prototype.getCurrent = function () {
        return this.current;
    };
    TransactionalObject.prototype.setProperty = function (path, changes) {
        var pathEntries = this.getPath(this.current, path);
        var lastEntry = pathEntries[pathEntries.length - 1];
        var parent = (lastEntry ? lastEntry.parent[lastEntry.field] : this.current);
        var newValue = this.merge(parent, changes);
        for (var i = pathEntries.length - 1; i >= 0; i--) {
            var entry = pathEntries[i];
            var parent_1 = entry.parent;
            var field = entry.field;
            newValue = this.setField(parent_1, field, newValue);
        }
        return this.current = newValue;
    };
    TransactionalObject.prototype.setField = function (parent, field, value) {
        if (parent[field] === value) {
            return parent;
        }
        var newParent = this.clone(parent);
        newParent[field] = value;
        return newParent;
    };
    TransactionalObject.prototype.getPath = function (obj, path) {
        var entries = [];
        if (path == ROOT) {
            return entries;
        }
        var parent = obj;
        var fields = path.split(".");
        for (var _i = 0, fields_1 = fields; _i < fields_1.length; _i++) {
            var field = fields_1[_i];
            if (parent === null || parent === undefined) {
                break;
            }
            if (typeof parent != "object") {
                throw new Error("Invalid path: " + path);
            }
            entries.push({ parent: parent, field: field });
            parent = parent[field];
        }
        return entries;
    };
    TransactionalObject.prototype.clone = function (obj) {
        if (obj && obj[$$MODIFIED]) {
            //
            //  This object is already a clone
            //  No need to clone again
            //
            return obj;
        }
        var res = Object.assign({}, obj);
        res[$$MODIFIED] = true;
        this.modified.push(res);
        res[$$VERSION] = (res[$$VERSION] || 0) + 1;
        return res;
    };
    TransactionalObject.prototype.merge = function (obj, changes) {
        if (obj == changes) {
            return changes;
        }
        if (typeof changes != "object") {
            //
            //  Cannot merge a value into itself
            //
            return changes;
        }
        if (Array.isArray(obj)) {
            //
            //  Cannot merge arrays
            //
            return changes;
        }
        var newObj = null;
        for (var field in changes) {
            if (changes.hasOwnProperty(field)) {
                var newValue = changes[field];
                var oldValue = obj[field];
                if (oldValue != newValue) {
                    newObj = newObj || this.clone(obj);
                    newObj[field] = newValue;
                }
            }
        }
        return newObj || obj;
    };
    return TransactionalObject;
}());
TransactionalObject.ROOT = ROOT;

var PathResolver = (function () {
    function PathResolver(path) {
        this.path = path;
        if (this.path == ROOT) {
            this.parts = [];
            return;
        }
        this.parts = this.path.split(".");
    }
    PathResolver.create = function (path) {
        return new PathResolver(path);
    };
    PathResolver.resolve = function (obj, path) {
        return new PathResolver(path).get(obj);
    };
    PathResolver.prototype.get = function (obj) {
        if (this.parts.length == 0) {
            return obj;
        }
        var res = obj;
        for (var _i = 0, _a = this.parts; _i < _a.length; _i++) {
            var part = _a[_i];
            res = res[part];
        }
        return res;
    };
    PathResolver.prototype.set = function (obj, val) {
        for (var i = 0; i < this.parts.length - 1; i++) {
            obj = obj[this.parts[i]];
        }
        obj[this.parts[this.parts.length - 1]] = val;
    };
    return PathResolver;
}());

var config = {
    enableLogging: false,
    activityAutoBeginTransaction: false,
    updateAutoBeginTransaction: true,
    allowConcurrencyErrors: true,
};
function configure(newConfig) {
    if (newConfig.enableLogging != config.enableLogging) {
        if (newConfig.enableLogging) {
            //Logger.root.enable();
        }
        else {
            //Logger.root.disable();
        }
    }
    Object.assign(config, newConfig);
}

var logger$1 = appLogger.create("AppStore");
//
//  Holds application base and allow subscribing to changes
//  Each ServiceStore register itself to it
//  Commit request is delegated from the ServiceStore to this appStore
//
var AppStore = (function () {
    function AppStore() {
        this.activityListeners = [];
        this.appState = {};
        this.listeners = [];
        this.stores = [];
        this.services = [];
        this.middleware = null;
        this.middlewareNext = null;
    }
    AppStore.prototype.init = function (services) {
        logger$1("BEGIN - init").log();
        for (var _i = 0, services_1 = services; _i < services_1.length; _i++) {
            var service = services_1[_i];
            this.registerStore(service.store);
        }
        var systemStore = ServiceStore.create("$$system", {
            initialized: false,
            nextActivityId: 1,
            nextTranId: 1,
        });
        this.registerStore(systemStore);
        this.services = services;
        logger$1("Registered services are", this.services).log();
        systemStore.update({
            initialized: true,
        });
        logger$1("Initial appState", this.appState).log();
        logger$1("END - init").log();
    };
    AppStore.prototype.executeReactions = function () {
        logger$1("BEGIN - executeReactions").log();
        for (var _i = 0, _a = this.services; _i < _a.length; _i++) {
            var service = _a[_i];
            var reactions = Reflect.get(service.constructor, "reactions");
            if (reactions) {
                for (var _b = 0, reactions_1 = reactions; _b < reactions_1.length; _b++) {
                    var reaction = reactions_1[_b];
                    var method = service[reaction];
                    if (method) {
                        logger$1("BEGIN - " + service.constructor.name + "." + reaction).log();
                        method.call(service);
                        logger$1("END - " + service.constructor.name + "." + reaction).log();
                    }
                }
            }
        }
        logger$1("END - executeReactions").log();
    };
    AppStore.prototype.extractStore = function (serviceOrStore) {
        if (serviceOrStore instanceof ServiceStore) {
            return serviceOrStore;
        }
        else if (serviceOrStore.store && serviceOrStore.store instanceof ServiceStore) {
            return serviceOrStore.store;
        }
        else {
            throw new Error("Invalid service/store instance");
        }
    };
    AppStore.prototype.getState = function () {
        return this.appState;
    };
    AppStore.prototype.subscribe = function (listener) {
        var _this = this;
        this.listeners.push(listener);
        return function () {
            _this.unsubscribe(listener);
        };
    };
    AppStore.prototype.unsubscribe = function (listener) {
        var index = this.listeners.indexOf(listener);
        if (index != -1) {
            this.listeners.splice(index, 1);
        }
    };
    AppStore.prototype.commit = function (oldState, newState) {
        if (newState == this.appState) {
            logger$1("Nothing new to commit").log();
            return;
        }
        if (!config.allowConcurrencyErrors && oldState != this.appState) {
            logger$1("A request for new state", newState, "with base", oldState, "is conflicting with existing state", this.appState);
            throw new Error("Concurrency error");
        }
        logger$1("Changing state", oldState, " ==> ", newState).log();
        this.appState = newState;
        //this.runReactions();
        //this.emit(oldState, newState);
        return this.appState;
    };
    AppStore.prototype.emit = function (oldState, newState) {
        for (var _i = 0, _a = this.listeners; _i < _a.length; _i++) {
            var l = _a[_i];
            try {
                l(this.appState, oldState);
            }
            catch (err) {
                logger$1("Ignoring error during AppStore change event", err).error();
            }
        }
    };
    AppStore.prototype.registerStore = function (store) {
        var metadata = store.getMetadata();
        var conflict = this.findConflictingPath(metadata.path);
        if (conflict) {
            throw new Error("Service with path: " + metadata.path + " conflicts with existing service with path: " + conflict.getMetadata().path);
        }
        this.stores.push(store);
        if (store.getMetadata().path == ROOT) {
            this.appState = store.getMetadata().initialState;
        }
        else {
            PathResolver.create(metadata.path).set(this.appState, metadata.initialState);
        }
        store.onRegistered(this);
    };
    AppStore.prototype.findConflictingPath = function (path) {
        if (path == ROOT) {
            return this.stores.find(function (s) { return s.getMetadata().path == ROOT; });
        }
        for (var _i = 0, _a = this.stores; _i < _a.length; _i++) {
            var store = _a[_i];
            if (path.indexOf(store.getMetadata().path) == 0) {
                return store;
            }
        }
        return null;
    };
    AppStore.prototype.registerListener = function (listener) {
        this.activityListeners.push(listener);
    };
    AppStore.prototype.unregisterListener = function (listener) {
        var index = this.activityListeners.indexOf(listener);
        if (index != -1) {
            this.listeners.splice(index, 1);
        }
    };
    AppStore.prototype.registerMiddleware = function (middleware) {
        var oldMiddleware = this.middleware;
        var oldMiddlewareNext = this.middlewareNext;
        this.middlewareNext = function (activity) {
            if (oldMiddleware) {
                return oldMiddleware(activity, oldMiddlewareNext);
            }
        };
        this.middleware = middleware;
    };
    AppStore.prototype.emitActivityEvent = function (func, activity) {
        logger$1("emitActivityEvent", activity).log();
        for (var _i = 0, _a = this.activityListeners; _i < _a.length; _i++) {
            var listener = _a[_i];
            func(listener);
        }
    };
    AppStore.prototype._onActivityBegin = function (activity) {
        this.emitActivityEvent(function (l) { return l.onActivityBegin(activity); }, activity);
    };
    AppStore.prototype._runActivity = function (activity, func) {
        if (this.middleware) {
            return this.middleware(activity, func);
        }
        return func();
    };
    AppStore.prototype._onActivityError = function (activity, err) {
        this.emitActivityEvent(function (l) { return l.onActivityError(activity, err); }, activity);
    };
    AppStore.prototype._onActivitySuccess = function (activity, res) {
        this.emitActivityEvent(function (l) { return l.onActivitySuccess(activity, res); }, activity);
    };
    AppStore.prototype._onActivitySyncComplete = function (activity) {
        this.emitActivityEvent(function (l) { return l.onActivitySyncComplete(activity); }, activity);
    };
    AppStore.prototype._onActivityAsyncComplete = function (activity) {
        this.emitActivityEvent(function (l) { return l.onActivityAsyncComplete(activity); }, activity);
    };
    AppStore.prototype._onActivityZoneComplete = function (activity) {
        this.emitActivityEvent(function (l) { return l.onActivityZoneComplete(activity); }, activity);
    };
    return AppStore;
}());

var COUNTERS = ["XMLHttpRequest.send", "setTimeout", "setInterval"];
var logger$4 = appLogger.create("ActivityScope");
var ActivityScope = (function () {
    function ActivityScope(appStore, name) {
        this.appStore = appStore;
        this.name = name;
        this.id = ++ActivityScope.nextActivityId;
        this.outerZone = Zone.current;
        this.counters = {};
        this.trans = 0;
    }
    ActivityScope.current = function () {
        var activity = Zone.current.get("activity");
        return activity;
    };
    ActivityScope.runInsideActivity = function (appStore, name, action) {
        logger$4("runInsideActivity", name).log();
        function runAction(func, isRoot, activity) {
            var retVal;
            try {
                retVal = func();
                if (isRoot) {
                    if (retVal && retVal.catch) {
                        activity.onSyncComplete();
                        retVal.then(function (res) {
                            activity.onSuccess(res);
                            activity.onAsyncComplete();
                        }, function (err) {
                            activity.onError(err);
                            activity.onAsyncComplete();
                        });
                    }
                    else {
                        activity.onSuccess(retVal);
                        activity.onSyncComplete();
                    }
                }
                return retVal;
            }
            catch (err) {
                if (isRoot) {
                    activity.onError(err);
                    activity.onSyncComplete();
                    return;
                }
                else {
                    throw err;
                }
            }
        }
        var activity = ActivityScope.current();
        if (activity) {
            logger$4("Nested activity. Parent is", activity).log();
            //
            //  This is a nested activity
            //  No need to catch errors
            //
            return runAction(action, false, activity);
        }
        //
        //  This is a root activity
        //  Need to monitor for errors and emit event when error has happened
        //
        activity = new ActivityScope(appStore, name);
        var spec = {
            name: "activity",
            properties: {
                "activity": activity,
            },
            onScheduleTask: function (parentZoneDelegate, currentZone, targetZone, task) {
                var activityScope = currentZone.get("activity");
                activityScope.onScheduleTask(task.source);
                return parentZoneDelegate.scheduleTask(targetZone, task);
            },
            onHasTask: function (parentZoneDelegate, currentZone, targetZone, hasTaskState) {
                if (!hasTaskState.macroTask && !hasTaskState.microTask) {
                    var activityScope = currentZone.get("activity");
                    activityScope.onZoneComplete();
                }
            }
        };
        var zone = Zone.current.fork(spec);
        return zone.run(function () {
            activity.onBegin();
            return activity.run(function () {
                return runAction(action, true, activity);
            });
        });
    };
    ActivityScope.prototype.onScheduleTask = function (source) {
        this.updateCounter(source);
    };
    ActivityScope.prototype.updateCounter = function (name) {
        if (COUNTERS.indexOf(name) == -1) {
            return;
        }
        var counter = this.counters[name];
        if (counter === undefined) {
            counter = 0;
        }
        ++counter;
        this.counters[name] = counter;
    };
    ActivityScope.prototype.onBegin = function () {
        logger$4("onBegin", this.name).log();
        this.startTime = new Date();
        this.appStore._onActivityBegin(this);
    };
    ActivityScope.prototype.run = function (func) {
        return this.appStore._runActivity(this, func);
    };
    ActivityScope.prototype.onSuccess = function (res) {
        logger$4("onSuccess", this.name, res).log();
        this.result = res;
        this.appStore._onActivitySuccess(this, res);
    };
    ActivityScope.prototype.onError = function (err) {
        logger$4("onError", err).log();
        this.appStore._onActivityError(this, err);
    };
    ActivityScope.prototype.onSyncComplete = function () {
        logger$4("onComplete", this.name, this.getStats()).log();
        this.endTime = new Date();
        this.appStore._onActivitySyncComplete(this);
    };
    ActivityScope.prototype.onAsyncComplete = function () {
        logger$4("onAsyncComplete", this.name, this.getStats()).log();
        this.endTimeAsync = new Date();
        this.appStore._onActivityAsyncComplete(this);
    };
    ActivityScope.prototype.onZoneComplete = function () {
        if (this.endTimeZone !== undefined) {
            return;
        }
        this.endTimeZone = new Date();
        logger$4("onZoneComplete", this.name, this.getStats()).log();
        this.appStore._onActivityZoneComplete(this);
    };
    ActivityScope.prototype.onTransactionCommitted = function () {
        ++this.trans;
    };
    ActivityScope.prototype.getStats = function () {
        return {
            result: this.result,
            time: (this.endTime - this.startTime),
            promiseTime: (this.endTimeAsync - this.startTime),
            zoneTime: (this.endTimeZone - this.startTime),
            trans: this.trans,
            counters: this.counters,
        };
    };
    return ActivityScope;
}());
ActivityScope.nextActivityId = 0;

if (typeof Zone === "undefined") {
    throw new Error("t-rex cannot execute without zone.js. Please ensure zone.js is loaded before t-rex");
}
var logger$3 = appLogger.create("TransactionScope");
var TransactionScope = (function () {
    function TransactionScope(appStore) {
        this.id = ++TransactionScope.nextTranId;
        this.updateCount = 0;
        this.appStore = appStore;
        this.oldState = appStore.getState();
        this.tranState = new TransactionalObject(this.oldState);
        this.committed = false;
        logger$3("created", this.oldState).log();
    }
    TransactionScope.prototype.update = function (path, changes) {
        logger$3("update(" + ++this.updateCount + ")", path, changes).log();
        this.ensureNotCommitted();
        var newState = this.tranState.setProperty(path, changes);
        if (newState != this.oldState) {
            logger$3("newState", newState).log();
        }
        else {
            logger$3("No effective changes").log();
        }
    };
    TransactionScope.prototype.getNewState = function () {
        return this.tranState.getCurrent();
    };
    TransactionScope.prototype.getOldState = function () {
        return this.tranState.getBase();
    };
    TransactionScope.prototype.commit = function () {
        this.ensureNotCommitted();
        var oldState = this.tranState.getBase();
        var newState = this.tranState.getCurrent();
        var currentState = this.appStore.getState();
        if (newState == currentState) {
            logger$3("Nothing new to commit. updateCount is", this.updateCount).log();
            return;
        }
        if (!config.allowConcurrencyErrors && oldState != currentState) {
            //
            //  A parallel transaction has already modified the main store before us
            //  We only allow parallel changes at different branches. Else, "Concurrency error is raised"
            //
            this.tranState.rebase(currentState);
            oldState = this.tranState.getBase();
            newState = this.tranState.getCurrent();
        }
        this.appStore.executeReactions();
        this.appStore.commit(oldState, newState);
        var activity = ActivityScope.current();
        if (activity) {
            activity.onTransactionCommitted();
        }
        this.committed = true;
        //
        //  Committing to appStore causes emitting of change event
        //  Subscribers must be notified outside of the transaction, else, any
        //  additional update will be considered as part of the already committed transaction
        //  and therefore will throw error
        //
        TransactionScope._current = null;
        this.appStore.emit(oldState, newState);
        //
        //  Cleans management flags + switch between old and new
        //
        this.tranState.commit();
        logger$3("committed").log();
    };
    Object.defineProperty(TransactionScope, "current", {
        get: function () {
            return TransactionScope._current;
        },
        enumerable: true,
        configurable: true
    });
    TransactionScope.runInsideTransaction = function (appStore, action) {
        if (TransactionScope.current) {
            return action();
        }
        var newTran = new TransactionScope(appStore);
        TransactionScope._current = newTran;
        try {
            var retVal = action();
            newTran.commit();
            return retVal;
        }
        finally {
            TransactionScope._current = null;
        }
    };
    TransactionScope.prototype.ensureNotCommitted = function () {
        if (this.committed) {
            throw new Error("Activity was already committed");
        }
    };
    return TransactionScope;
}());
TransactionScope.nextTranId = 0;
TransactionScope._current = null;

function P1(obj, key1) {
    if (!obj) {
        return undefined;
    }
    return obj[key1]; // Inferred type is T[K]
}

function P2(obj, key1, key2) {
    return P1(P1(obj, key1), key2);
}



// export const promisify = (function() {
//     if(typeof window != "undefined" && window.hasOwnProperty("angular")) {
//         const angular = window["angular"];
//         let injector;
//         let $q;
//
//         return function promisifyOnAngular1(value) {
//             injector = injector || angular.element(document.getElementsByTagName("body")).injector();
//             $q = $q || injector.get("$q");
//
//             if(value && value.then) {
//                 return value;
//             }
//             return $q.when(value);
//         }
//     }
//     else {
//         return function promisifyOnNonAngular1(value) {
//             if(value && value.then) {
//                 return value;
//             }
//
//             return Promise.resolve(value);
//         }
//     }
// })();
//

var logger$5 = appLogger.create("operators");
var StoreOperator = (function () {
    function StoreOperator() {
    }
    return StoreOperator;
}());
var PushOperator = (function (_super) {
    __extends(PushOperator, _super);
    function PushOperator(item) {
        var _this = _super.call(this) || this;
        _this.item = item;
        return _this;
    }
    PushOperator.prototype.execute = function (data) {
        logger$5("PushOperator.execute", data, this.item).log();
        var newData = data.concat([]);
        newData.push(this.item);
        return newData;
    };
    return PushOperator;
}(StoreOperator));
var IncrementOperator = (function (_super) {
    __extends(IncrementOperator, _super);
    function IncrementOperator() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    IncrementOperator.prototype.execute = function (data) {
        logger$5("IncrementOperator.execute", data).log();
        return data + 1;
    };
    return IncrementOperator;
}(StoreOperator));
var DecrementOperator = (function (_super) {
    __extends(DecrementOperator, _super);
    function DecrementOperator() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    DecrementOperator.prototype.execute = function (data) {
        logger$5("DecrementOperator.execute", data).log();
        return data - 1;
    };
    return DecrementOperator;
}(StoreOperator));
function push(item) {
    return new PushOperator(item);
}
function dec() {
    return new DecrementOperator();
}
function inc() {
    return new IncrementOperator();
}

var logger = appLogger.create("ServiceStore");
var ServiceStore = (function () {
    function ServiceStore(metadata) {
        this.appStore = null;
        this.metadata = metadata;
        this.pathResolver = new PathResolver(this.metadata.path);
        this.listeners = [];
    }
    ServiceStore.prototype.runWithOwnAppStore = function () {
        logger("runWithOwnAppStore").log();
        var appStore = new AppStore();
        appStore.init([
            { store: this }
        ]);
    };
    ServiceStore.prototype.onRegistered = function (appStore) {
        var _this = this;
        this.appStore = appStore;
        appStore.subscribe(function (newAppState, oldAppState) {
            var newState = _this.pathResolver.get(newAppState);
            var oldState = _this.pathResolver.get(oldAppState);
            if (oldState != newState) {
                _this.emit(newState, oldState);
            }
        });
    };
    ServiceStore.prototype.subscribe = function (listener) {
        var _this = this;
        this.ensureInitialized();
        this.listeners.push(listener);
        //
        //  Always notify new subscribers of the existing data
        //
        var state = this.pathResolver.get(this.appStore.getState());
        listener(state, state);
        return function () {
            _this.unsubscribe(listener);
        };
    };
    ServiceStore.prototype.unsubscribe = function (listener) {
        var index = this.listeners.indexOf(listener);
        if (index != -1) {
            this.listeners.splice(index, 1);
        }
    };
    ServiceStore.prototype.subscribeTo = function (path, listener) {
        this.ensureInitialized();
        var pathResolver = new PathResolver(path);
        this.listeners.push(function (s1, s2) {
            var c1 = pathResolver.get(s1);
            var c2 = pathResolver.get(s2);
            if (c1 != c2) {
                listener(c1, c2);
            }
        });
    };
    ServiceStore.prototype.subscribeTo1 = function (key1, listener) {
        this.ensureInitialized();
        this.listeners.push(function (s1, s2) {
            var c1 = P1(s1, key1);
            var c2 = P1(s2, key1);
            if (c1 != c2) {
                listener(c1, c2);
            }
        });
    };
    ServiceStore.prototype.subscribeTo2 = function (key1, key2, listener) {
        this.ensureInitialized();
        this.listeners.push(function (s1, s2) {
            var c1 = P2(s1, key1, key2);
            var c2 = P2(s2, key1, key2);
            if (c1 != c2) {
                listener(c1, c2);
            }
        });
    };
    ServiceStore.prototype.getAppStore = function () {
        this.ensureInitialized();
        return this.appStore;
    };
    ServiceStore.prototype.getMetadata = function () {
        return this.metadata;
    };
    ServiceStore.prototype.getState = function () {
        this.ensureInitialized();
        var tranScope = TransactionScope.current;
        var appState = (tranScope ? tranScope.getNewState() : this.appStore.getState());
        var state = this.pathResolver.get(appState);
        return state;
    };
    ServiceStore.prototype.resolveUpdateOverloading = function (key, value) {
        var changes;
        if (arguments.length == 2) {
            changes = {};
            changes[key] = value;
        }
        else if (typeof key == "object") {
            changes = key;
        }
        else {
            changes = key(this.getState());
        }
        return changes;
    };
    ServiceStore.prototype.update = function (key, value) {
        var _this = this;
        var changes = this.resolveUpdateOverloading.apply(this, arguments);
        logger("update", changes).log();
        this.ensureInitialized();
        var updated;
        if (config.updateAutoBeginTransaction) {
            TransactionScope.runInsideTransaction(this.appStore, function () {
                _this.doUpdate(TransactionScope.current, changes);
            });
            updated = this.getState();
        }
        else {
            var tranScope = TransactionScope.current;
            if (!tranScope) {
                throw new Error("No ambient transaction to update");
            }
            this.doUpdate(TransactionScope.current, changes);
            updated = this.getState();
        }
        if (arguments.length == 2) {
            return updated[key];
        }
        return updated;
    };
    ServiceStore.prototype.resolveOperators = function (changes, oldState) {
        var resolved = {};
        for (var key in changes) {
            var value = changes[key];
            if (value instanceof StoreOperator) {
                resolved[key] = value.execute(oldState[key]);
            }
            else {
                resolved[key] = value;
            }
        }
        return resolved;
    };
    ServiceStore.prototype.doUpdate = function (tranScope, changes) {
        var oldState = this.getState();
        var resolvedChanges = this.resolveOperators(changes, oldState);
        tranScope.update(this.metadata.path, resolvedChanges);
    };
    ServiceStore.create = function (path, initialState) {
        return new ServiceStore({
            path: path,
            initialState: initialState,
        });
    };
    ServiceStore.prototype.emit = function (newState, oldState) {
        for (var _i = 0, _a = this.listeners; _i < _a.length; _i++) {
            var l = _a[_i];
            try {
                l(newState, oldState);
            }
            catch (err) {
                logger("Ignoring error during ServiceStore change event", err).error();
            }
        }
    };
    ServiceStore.prototype.ensureInitialized = function () {
        if (!this.appStore) {
            throw new Error("Store is not initialized. Did you call appStore.init ?");
        }
    };
    return ServiceStore;
}());

var logger$6 = appLogger.create("decorators");
function transaction(store, func) {
    var appStore = getAppStoreFromStore(store);
    return TransactionScope.runInsideTransaction(appStore, function () {
        return func();
    });
}
function getAppStoreFromStore(store) {
    if (store instanceof AppStore) {
        return store;
    }
    if (store instanceof ServiceStore) {
        return store.getAppStore();
    }
    throw new Error("Unexpected store value. Should be of type AppStore | ServiceStore");
}
function getAppStoreFromService(service) {
    var appStore;
    if (service.store) {
        if (!(service.store instanceof ServiceStore)) {
            throw new Error("Unexpected store value. Should be of type ServiceStore");
        }
        return service.store.getAppStore();
    }
    else if (service.appStore) {
        if (!(service.appStore instanceof AppStore)) {
            throw new Error("Unexpected appStore value. Should be of type AppStore");
        }
        return service.appStore;
    }
    throw new Error("No store/appStore field was found for service instance");
}

function getOptionOrDefault(options, propName, defValue) {
    var propVal;
    if (options && options.hasOwnProperty(propName)) {
        propVal = options[propName];
    }
    else {
        propVal = defValue;
    }
    return propVal;
}
function Activity(options) {
    return function (target, propertyKey, descriptor) {
        var method = descriptor.value;
        descriptor.value = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var service = this;
            var appStore = getAppStoreFromService(service);
            var beginTransaction = getOptionOrDefault(options, "beginTransaction", config.activityAutoBeginTransaction);
            return ActivityScope.runInsideActivity(appStore, service.constructor.name + ":" + propertyKey, function () {
                if (beginTransaction) {
                    return TransactionScope.runInsideTransaction(appStore, function () {
                        return method.apply(service, args);
                    });
                }
                else {
                    return method.apply(service, args);
                }
            });
        };
        return descriptor;
    };
}
function Reaction() {
    return function (target, propertyKey, descriptor) {
        var ctor = target.constructor;
        if (!ctor) {
            logger$6("Target of @Reaction has no constructor").warn();
            return;
        }
        var reactions = Reflect.get(ctor, "reactions");
        if (!reactions) {
            reactions = [];
            Reflect.set(ctor, "reactions", reactions);
        }
        reactions.push(propertyKey);
        return descriptor;
    };
}

//Logger.root.name = "[t-rex]";

export { ServiceStore, AppStore, Activity, transaction, Reaction, TransactionScope, configure, ActivityScope, push, dec, inc };
